﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using Microsoft.AspNetCore.Mvc;
using System.Web;
using System.Net.Mime;
using GcPdfViewerSupportApi.Utils;
using GcPdfViewerSupportApi.Models;
using System.Security.Cryptography.X509Certificates;

namespace GcPdfViewerSupportApi.Controllers
{
    /// <summary>
    /// GcPdfViewer Support API controller.
    /// </summary>
    /// <example>
    /// Configure controller using ASP.NET Core Application Parts:
    /// <code>
    ///    public void ConfigureServices(IServiceCollection services)
    ///    {
    ///      services.AddMvc().ConfigureApplicationPartManager(apm => 
    ///         apm.ApplicationParts.Add(new AssemblyPart(typeof(GcPdfViewerController).GetTypeInfo().Assembly)));
    ///      ...
    /// </code>
    /// </example>
    [Route("SupportApi/GcPdfViewer")]
    [ApiController]
    [IgnoreAntiforgeryToken]
    public class GcPdfViewerController : ControllerBase
    {

        protected static Dictionary<string, GcPdfDocumentLoader> _docLoaders = new Dictionary<string, GcPdfDocumentLoader>();
        protected static Dictionary<string, DateTime> _docLoaderLastPingTime = new Dictionary<string, DateTime>();

        /// <summary>
        /// Global Support API settings.
        /// </summary>
        public static SupportApiSettings Settings { get; } = new SupportApiSettings();


        #region ** Web API

        /// <summary>
        /// Returns Support API version. 
        /// </summary>
        /// <returns>Support API version</returns>
        /// <example>
        /// http://localhost:50016/SupportApi/GcPdfViewer/Version
        /// </example>
        [HttpGet("Version")]
        public virtual string Version()
        {
            return GetType().Assembly.GetName().Version.ToString();
        }

        /// <summary>
        /// Open PDF document using specified url.
        /// </summary>
        /// <param name="url">Encoded url. Encoding is done using client side encodeURIComponent function</param>
        /// <returns>Information about opened document.</returns>      
        [Route("OpenUrl/{url}")]
        public virtual OpenedDocumentInfo OpenUrl(string url)
        {
            url = HttpUtility.UrlDecode(url);
            if (!string.IsNullOrEmpty(url) && !url.StartsWith("http"))
            {
                url = $"{Request.Scheme}://{(url.StartsWith(Request.Host.Value) ? "" : Request.Host.Value)}/{(Request.PathBase.HasValue ? Request.PathBase.Value : "")}{url}";
            }
            return _CreateDocLoader(url).Info;
        }

        /// <summary>
        /// Open PDF document using binary data.
        /// </summary>
        /// <returns>Information about opened document</returns>
        [Route("OpenBinary")]
        [DisableRequestSizeLimit]
        [RequestFormLimits(MultipartBodyLengthLimit = 268435456, KeyLengthLimit = 268435456, ValueLengthLimit = 268435456, ValueCountLimit = 268435456)]
        public virtual async System.Threading.Tasks.Task<OpenedDocumentInfo> OpenBinaryAsync()
        {
            var ms = new MemoryStream();
            await Request.Body.CopyToAsync(ms);        
            var loader = _CreateDocLoader(ms);
            return loader.Info;
        }

        [Route("Sign")]
        public virtual OpenedDocumentInfo Sign(string docId, [FromBody] SignatureInfo signatureInfo)
        {
            var docLoader = _GetDocLoader(docId);
            
            var certStream = new MemoryStream();
            Request.Body.CopyTo(certStream);
            
            docLoader.Sign(signatureInfo);

            return docLoader.Info;
        }

        /// <summary>
        /// Verify Signature field.
        /// </summary>
        /// <param name="docId">Document Identifier</param>
        /// <param name="fieldName">Signature field name</param>
        /// <returns></returns>
        [Route("VerifySignature/{docId}/{fieldName}")]
        public virtual IActionResult VerifySignature(string docId, string fieldName)
        {
            fieldName = HttpUtility.UrlDecode(fieldName);
            var docLoader = _GetDocLoader(docId);
            if (docLoader.VerifySignature(fieldName))
            {
                return Content("true");
            }
            return Content("false");
        }

        /// <summary>
        /// Close opened document and free server resources.
        /// </summary>
        /// <param name="docId">Document Identifier</param>
        /// <returns></returns>
        [Route("Close/{docId}")]
        public virtual string Close(string docId)
        {
            _DisposeDocLoader(docId);
            return "ok";

        }

        /// <summary>
        /// Ping from pdf viewer.
        /// </summary>
        /// <param name="docId">Optional. Document Identifier</param>
        /// <returns></returns>
        [Route("Ping/{docId?}")]
        public virtual string Ping(string docId)
        {
            if (!string.IsNullOrEmpty(docId) && _docLoaderLastPingTime.ContainsKey(docId))
            {
                _docLoaderLastPingTime[docId] = DateTime.Now;
                _DisposeInactiveDocuments();
                return "ok";
            }
            _DisposeInactiveDocuments();
            return "not-found";

        }

        /// <summary>
        /// Modify opened document.
        /// </summary>
        /// <param name="docId">Document Identifier</param>
        /// <param name="modifications">Document modifications</param>
        /// <returns></returns>
        [Route("Modify/{docId}")]
        public virtual async System.Threading.Tasks.Task<string> ModifyAsync(string  docId)
        {
            var ms = new MemoryStream();
            Stream body = Request.Body;
            await body.CopyToAsync(ms);
            var bytes = ms.ToArray();
            string json = System.Text.Encoding.UTF8.GetString(bytes);
            DocumentModifications modifications = Newtonsoft.Json.JsonConvert.DeserializeObject<DocumentModifications>(json);
            var docLoader = _GetDocLoader(docId);
            docLoader.ApplyDocumentModifications(modifications);
            return "ok";
        }

        /// <summary>
        /// Discard all changes made by Modify method.
        /// </summary>
        /// <param name="docId"></param>
        /// <returns></returns>
        [Route("Reset/{docId}")]
        public virtual string Reset(string docId)
        {
            var docLoader = _GetDocLoader(docId);
            docLoader.Reset();
            return "ok";

        }
      
        /// <summary>
        /// Apply modifications and download PDF document as file.
        /// </summary>
        /// <param name="docId"></param>
        /// <param name="fileName"></param>
        /// <returns></returns>
        [HttpGet("Download/{docId}/{fileName}")]
        public virtual IActionResult Download(string docId, string fileName)
        {
            var docLoader = _GetDocLoader(docId);
            var content = new MemoryStream();
            docLoader.Save(content);
            fileName = string.IsNullOrEmpty(fileName) ? $"{docLoader.DocId}.pdf" : fileName;
            ContentDisposition contentDisposition = new ContentDisposition() {
                FileName = fileName,
                Inline = false,
                Size = content.Length
            };
            Response.Headers["Content-Disposition"] = contentDisposition.ToString();
            var contentType = "application/octet-stream";
            return new FileStreamResult(content, contentType);
        }

        #endregion

        #region ** private implementations

        private GcPdfDocumentLoader _CreateDocLoader(Stream data)
        {
            var loader = new GcPdfDocumentLoader(Settings);
            _docLoaders[loader.DocId] = loader;
            _docLoaderLastPingTime[loader.DocId] = DateTime.Now;
            loader.Open(data);
            return loader;
        }

        private GcPdfDocumentLoader _CreateDocLoader(string url)
        {
            var loader = (from k in _docLoaders.Values where k.Url == url select k).FirstOrDefault();
            if (loader == null)
            {
                loader = new GcPdfDocumentLoader(Settings);
                _docLoaders[loader.DocId] = loader;
                _docLoaderLastPingTime[loader.DocId] = DateTime.Now;
                loader.Open(url);
            }
            return loader;
        }

        private void _DisposeInactiveDocuments()
        {
            var dt = DateTime.Now.AddMinutes(-10);
            var inactiveDocIds = _docLoaderLastPingTime.Where(kv => kv.Value < dt).Select(kv => kv.Key).ToArray();
            foreach (var docId in inactiveDocIds)
            {
                _DisposeDocLoader(docId);
            }
        }

        private void _DisposeDocLoader(string docId)
        {
            if(_docLoaderLastPingTime.ContainsKey(docId))
                _docLoaderLastPingTime.Remove(docId);
            var loader = _docLoaders.ContainsKey(docId) ? _docLoaders[docId] : null;
            if (loader == null)
            {
                return;
            }
            loader.Dispose();
            _docLoaders.Remove(docId);
        }

        private GcPdfDocumentLoader _GetDocLoader(string docId)
        {
            
            var loader = _docLoaders.ContainsKey(docId) ? _docLoaders[docId] : null;
            if (loader == null)
            {
                throw new Exception($"Document loader for document with id {docId} not found.");
            }
            return loader;
        }

        #endregion

    }

}
