﻿using GcPdfViewerSupportApi.Models;
using GrapeCity.Documents.Pdf;
using GrapeCity.Documents.Pdf.AcroForms;
using GrapeCity.Documents.Pdf.Actions;
using GrapeCity.Documents.Pdf.Annotations;
using Newtonsoft.Json.Linq;
using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Text;

namespace GcPdfViewerSupportApi.Utils
{
    public class AcroFormFactory 
    {
        #region ** public methods

        public static Field CreateField(DataParser dataParser, Page page)
        {
            AnnotationTypeCode annotationType = dataParser.annotationType;
            switch (annotationType)
            {
                case AnnotationTypeCode.WIDGET:
                    string fieldType = dataParser.ParseString("fieldType");
                    
                    if (fieldType == "Tx")
                    {
                        // Text field widgets
                        bool comb = dataParser.ParseBool("comb");
                        bool multiLine = dataParser.ParseBool("multiLine");
                        bool hasPasswordFlag = dataParser.ParseBool("hasPasswordFlag");
                        if (comb)
                        {
                            return UpdateField(new CombTextField(), dataParser);
                        }
                        else if (multiLine)
                        {
                            return UpdateField(new TextField() { Multiline = true }, dataParser);
                        }
                        else if (hasPasswordFlag)
                        {
                            return UpdateField(new TextField() { Password = true }, dataParser);
                        }
                        else
                        {
                            return UpdateField(new TextField(), dataParser);
                        }
                    }
                    else if (fieldType == "Btn")
                    {
                        // Buttons widgets
                        bool checkBox = dataParser.ParseBool("checkBox");
                        bool radioButton = dataParser.ParseBool("radioButton");
                        bool pushButton = dataParser.ParseBool("pushButton");
                        if (checkBox)
                        {
                            return UpdateField(new CheckBoxField(), dataParser);
                        }
                        else if (radioButton)
                        {
                            var radioButtonField = new RadioButtonField();
                            var widget = new WidgetAnnotation(dataParser.page, dataParser.rect);
                            radioButtonField.Widgets.Add(widget);
                            UpdateFieldWidget(radioButtonField, widget, dataParser);
                            return radioButtonField;

                        }
                        else if (pushButton)
                        {
                            return UpdateField(new PushButtonField(), dataParser);
                        }
                        else
                        {
                            return UpdateField(new PushButtonField(), dataParser);
                        }
                    }
                    else if (fieldType == "Ch")
                    {
                        // Choice widgets
                        bool combo = dataParser.ParseBool("combo");
                        bool multiSelect = dataParser.ParseBool("multiSelect");
                        if (combo)
                        {
                            return UpdateField(new ComboBoxField(), dataParser);
                        }
                        else if (multiSelect)
                        {
                            return UpdateField(new ListBoxField(), dataParser);
                        }
                        else
                        {
                            throw new Exception("Unknown choice field");
                        }
                    }
                    return null;
                case AnnotationTypeCode.SIGNATURE:
                    return UpdateField(new SignatureField(), dataParser);
                default:
                    return null;
            }
        }

        public static void UpdateField(AnnotationTypeCode annotationType, Field field, DataParser dataParser)
        {
            switch (annotationType)
            {
                case AnnotationTypeCode.WIDGET:
                    string fieldType = dataParser.ParseString("fieldType");

                    if (fieldType == "Tx")
                    {
                        // Text field widgets
                        bool comb = dataParser.ParseBool("comb");
                        bool multiLine = dataParser.ParseBool("multiLine");
                        bool hasPasswordFlag = dataParser.ParseBool("hasPasswordFlag");
                        if (comb)
                        {
                            UpdateField(field as CombTextField, dataParser);
                        }
                        else if (multiLine)
                        {
                            UpdateField(field as TextField, dataParser);
                        }
                        else if (hasPasswordFlag)
                        {
                            UpdateField(field as TextField, dataParser);
                        }
                        else
                        {
                            UpdateField(field as TextField, dataParser);
                        }
                    }
                    else if (fieldType == "Btn")
                    {
                        // Buttons widgets
                        bool checkBox = dataParser.ParseBool("checkBox");
                        bool radioButton = dataParser.ParseBool("radioButton");

                        bool pushButton = dataParser.ParseBool("pushButton");
                        if (checkBox)
                        {
                            UpdateField(field as CheckBoxField, dataParser);
                        }
                        else if (radioButton)
                        {
                            UpdateField(field as RadioButtonField, dataParser);
                        }
                        else if (pushButton)
                        {
                            UpdateField(field as PushButtonField, dataParser);
                        }
                        else
                        {
                            UpdateField(field as PushButtonField, dataParser);
                        }
                    }
                    else if (fieldType == "Ch")
                    {
                        // Choice widgets
                        bool combo = dataParser.ParseBool("combo");
                        bool multiSelect = dataParser.ParseBool("multiSelect");
                        if (combo)
                        {
                            UpdateField(field as ComboBoxField, dataParser);
                        }
                        else if (multiSelect)
                        {
                            UpdateField(field as ListBoxField, dataParser);
                        }
                        else
                        {
                            throw new Exception("Unknown choice field");
                        }
                    }
                    break;
                case AnnotationTypeCode.SIGNATURE:
                    UpdateField(field as SignatureField, dataParser);
                    break;
                default:
                    break;
            }
        }

        #endregion

        #region ** private implementation

        public static void UpdateFieldWidget(Field field, WidgetAnnotation widget, DataParser dataParser)
        {
            if (widget == null)
            {
                widget = new WidgetAnnotation(dataParser.page, dataParser.rect);
                field.Widgets.Add(widget);
            }
            else
            {
                widget.Page = dataParser.page;
                widget.Rect = dataParser.rect;
            }

            if (!string.IsNullOrEmpty(dataParser.name))
                widget.Name = dataParser.name;
            if (!string.IsNullOrEmpty(dataParser.FieldName))
            {
                field.Name = dataParser.FieldName;
            }
            if (dataParser.annotationData.ContainsKey("backgroundColor"))
            {
                widget.BackColor = dataParser.ParseColor("backgroundColor", Color.Transparent).Value;
            }
            Color appearanceColor = dataParser.ParseColor("appearanceColor", Color.Empty).Value;
            if (dataParser.annotationData.ContainsKey("borderColor"))
            {
                widget.Border.Color = dataParser.ParseColor("borderColor", Color.Transparent).Value;
            }
            else
            {
                if (!appearanceColor.IsEmpty)
                    widget.Border.Color = appearanceColor;
            }
            if (!appearanceColor.IsEmpty)
                widget.TextFormat.ForeColor = appearanceColor;
            widget.Contents = dataParser.contents;
            if (dataParser.borderWidth.HasValue)
            {
                widget.Border.Width = dataParser.borderWidth.Value;
            }
            if (dataParser.borderStyleType.HasValue)
            {
                widget.Border.Style = dataParser.borderStyleType.Value;
            }
            if (dataParser.annotationData.ContainsKey("fontSize"))
            {
                int fontSize = dataParser.ParseNumber("fontSize", 12);
                widget.TextFormat.FontSize = fontSize;
            }


            string fieldName = dataParser.FieldName;
            if (string.IsNullOrEmpty(fieldName))
            {
                fieldName = $"autogenerated_name_{Guid.NewGuid().ToString()}";
            }
            field.Name = fieldName;

            string fieldValue = dataParser.ParseString("fieldValue");
            field.Value = fieldValue;
            if (field is RadioButtonField)
            {
                var buttonName = dataParser.ParseString("buttonValue");
                widget.Name = buttonName;
            }
            else if (field is PushButtonField)
            {
                widget.ButtonAppearance.Caption = fieldValue;
            }
            else if (field is CheckBoxField)
            {
                (field as CheckBoxField).Value = dataParser.ParseBool("fieldValue");
            }

            field.ReadOnly = dataParser.ParseBool("readOnly");
            if (Enum.TryParse(dataParser.ParseString("textAlignment"), out VariableTextJustification justification))
            {
                field.Justification = justification;
            }
        }

        private static Field UpdateField(SignatureField field, DataParser dataParser)
        {
            UpdateFieldWidget(field, field.Widget, dataParser);
            return field;
        }

        private static Field UpdateField(CombTextField field, DataParser dataParser)
        {
            UpdateField((TextField)field, dataParser);
            return field;
        }

        private static Field UpdateField(TextField field, DataParser dataParser)
        {
            UpdateFieldWidget(field, field.Widget, dataParser);
            if (dataParser.annotationData.ContainsKey("maxLen"))
                field.MaxLen = dataParser.ParseNumber("maxLen");
            return field;
        }

        private static Field UpdateField(CheckBoxField field, DataParser dataParser)
        {
            UpdateFieldWidget(field, field.Widget, dataParser);
            return field;
        }

        private static Field UpdateField(RadioButtonField field, DataParser dataParser)
        {
            var widget = field.Widgets.Where(w => w.ParsedObjID == dataParser.pdfObjID).FirstOrDefault();
            UpdateFieldWidget(field, widget, dataParser);
            return field;
        }

        private static Field UpdateField(PushButtonField field, DataParser dataParser)
        {
            bool SubmitForm = dataParser.ParseBool("SubmitForm");
            bool ResetForm = dataParser.ParseBool("ResetForm");
            UpdateFieldWidget(field, field.Widget, dataParser);
            if (ResetForm)
            {
                field.Widget.Events.Activate = new ActionResetForm();
            }
            else if (SubmitForm)
            {
                string submitUrl = dataParser.ParseString("submitUrl");
                field.Widget.Events.Activate = new ActionSubmitForm(submitUrl);
            }
            return field;
        }

        private static Field UpdateField(ComboBoxField field, DataParser dataParser)
        {
            UpdateFieldWidget(field, field.Widget, dataParser);
            _ReadItems(field, dataParser);
            return field;
        }

        private static Field UpdateField(ListBoxField field, DataParser dataParser)
        {
            UpdateFieldWidget(field, field.Widget, dataParser);
            _ReadItems(field, dataParser);
            return field;
        }

        private static void _ReadItems(ChoiceField choiceField, DataParser dataParser)
        {
            var items = choiceField.Items;
            items.Clear();
            if (dataParser.TryParseArray("options", out JToken[] array))
            {
                foreach (JToken token in array)
                {
                    var obj = token as JObject;
                    items.Add(new ChoiceFieldItem(dataParser.ParseString(obj, "displayValue", ""), dataParser.ParseString(obj, "exportValue", "")));
                }
            }
        }

        #endregion

    }
}
