﻿using System;
using System.Collections.Generic;
using System.Drawing;
using GrapeCity.Documents.Pdf;
using GrapeCity.Documents.Pdf.Annotations;
using GcPdfViewerSupportApi.Models;
using Newtonsoft.Json.Linq;

namespace GcPdfViewerSupportApi.Utils
{

    /// <summary>
    /// AnnotationFactory class.
    /// </summary>
    public class AnnotationFactory
    {
        #region ** public methods

        /// <summary>
        /// Create annotation.
        /// </summary>
        /// <param name="annotationData"></param>
        /// <param name="page"></param>
        /// <returns></returns>
        public static AnnotationBase CreateAnnotation(DataParser dataParser, Page page)
        {
            AnnotationTypeCode annotationType = dataParser.annotationType;
            switch (annotationType)
            {
                case AnnotationTypeCode.CIRCLE:
                    return UpdateAnnotation(new CircleAnnotation(), dataParser);
                case AnnotationTypeCode.LINE:
                    return UpdateAnnotation(new LineAnnotation(), dataParser);
                case AnnotationTypeCode.SQUARE:
                    return UpdateAnnotation(new SquareAnnotation(), dataParser);
                case AnnotationTypeCode.REDACT:
                    return UpdateAnnotation(new RedactAnnotation(), dataParser);
                case AnnotationTypeCode.FREETEXT:
                    return UpdateAnnotation(new FreeTextAnnotation(), dataParser);                    
                case AnnotationTypeCode.TEXT:
                    return UpdateAnnotation(new TextAnnotation(), dataParser);
                case AnnotationTypeCode.INK:
                    return UpdateAnnotation(new InkAnnotation(), dataParser);
                case AnnotationTypeCode.POLYLINE:
                    return UpdateAnnotation(new PolyLineAnnotation(), dataParser);
                case AnnotationTypeCode.POLYGON:
                    return UpdateAnnotation(new PolygonAnnotation(), dataParser);
                case AnnotationTypeCode.FILEATTACHMENT:
                    return UpdateAnnotation(new FileAttachmentAnnotation(), dataParser);
                case AnnotationTypeCode.SOUND:
                    return UpdateAnnotation(new SoundAnnotation(), dataParser);
                case AnnotationTypeCode.POPUP:
                    return UpdateAnnotation(new PopupAnnotation(), dataParser);
            }            
            return null;
        }

        /// <summary>
        /// Update annotation.
        /// </summary>
        /// <param name="annotation"></param>
        /// <param name="annotationData"></param>
        /// <param name="page"></param>
        public static void UpdateAnnotation(AnnotationTypeCode annotationType, AnnotationBase annotation, DataParser dataParser)
        {
            switch (annotationType)
            {
                case AnnotationTypeCode.CIRCLE:
                    UpdateAnnotation((CircleAnnotation)annotation, dataParser);
                    break;
                case AnnotationTypeCode.LINE:
                    UpdateAnnotation((LineAnnotation)annotation, dataParser);
                    break;
                case AnnotationTypeCode.SQUARE:
                    UpdateAnnotation((SquareAnnotation)annotation, dataParser);
                    break;
                case AnnotationTypeCode.REDACT:
                    UpdateAnnotation((RedactAnnotation)annotation, dataParser);
                    break;
                case AnnotationTypeCode.FREETEXT:
                    UpdateAnnotation((FreeTextAnnotation)annotation, dataParser);
                    break;
                case AnnotationTypeCode.TEXT:
                    UpdateAnnotation((TextAnnotation)annotation, dataParser);
                    break;
                case AnnotationTypeCode.INK:
                    UpdateAnnotation((InkAnnotation)annotation, dataParser);
                    break;
                case AnnotationTypeCode.POLYLINE:
                    UpdateAnnotation((PolyLineAnnotation)annotation, dataParser);
                    break;
                case AnnotationTypeCode.POLYGON:
                    UpdateAnnotation((PolygonAnnotation)annotation, dataParser);
                    break;
                case AnnotationTypeCode.FILEATTACHMENT:
                    UpdateAnnotation((FileAttachmentAnnotation)annotation, dataParser);
                    break;
                case AnnotationTypeCode.SOUND:
                    UpdateAnnotation((SoundAnnotation)annotation, dataParser);
                    break;
                case AnnotationTypeCode.POPUP:
                    UpdateAnnotation((PopupAnnotation)annotation, dataParser);
                    break;
            }
        }

        #endregion

        #region ** private implementation

        private static void _UpdateMarkupAnnotation(MarkupAnnotation annotation, DataParser dataParser)
        {            
            annotation.Rect = dataParser.rect;
            if (!string.IsNullOrEmpty(dataParser.subject))
                annotation.Subject = dataParser.subject;
            if (!string.IsNullOrEmpty(dataParser.name))
                annotation.Name = dataParser.name;
            if (!dataParser.color.IsEmpty)
                annotation.Color = dataParser.color;
            if (dataParser.isRichContents)
                annotation.RichText = dataParser.contents;
            else
                annotation.Contents = dataParser.contents;
            if (!string.IsNullOrEmpty(dataParser.title))
                annotation.UserName = dataParser.title;
        }

        private static AnnotationBase UpdateAnnotation(PopupAnnotation annotation, DataParser dataParser)
        {
            annotation.Rect = dataParser.rect;
            if (!string.IsNullOrEmpty(dataParser.name))
                annotation.Name = dataParser.name;
            if (!dataParser.color.IsEmpty)
                annotation.Color = dataParser.color;
            annotation.Contents = dataParser.contents;
            annotation.Open = dataParser.open;            
            return annotation;
        }

        private static AnnotationBase UpdateAnnotation(FileAttachmentAnnotation annotation, DataParser dataParser)
        {
            _UpdateMarkupAnnotation(annotation, dataParser);
            var data = dataParser.annotationData;
            bool remove_content = dataParser.ParseBool("remove_content", false);
            if(remove_content)
            {
                annotation.File = null;
            }
            if (data.ContainsKey("file"))
            {
                if (data.TryGetValue("file", out var fileDataDict))
                {
                    string filename = fileDataDict.ContainsKey("filename") ? fileDataDict["filename"].ToString() : string.Empty;
                    string file_encoded_content = data.ContainsKey("file_encoded_content") ? data["file_encoded_content"].ToString() : string.Empty;
                    if (!string.IsNullOrEmpty(file_encoded_content))
                    {
                        byte[] bytes = System.Convert.FromBase64String(file_encoded_content);
                        annotation.File = FileSpecification.FromEmbeddedStream(filename, EmbeddedFileStream.FromBytes(dataParser.page.Doc, bytes));
                    }
                    else
                    {
                        if (annotation.File != null)
                            annotation.File.Desc = filename;
                    }
                }
            }
            if (!string.IsNullOrEmpty(dataParser.name))
            {
                annotation.Name = dataParser.name;
                if (Enum.TryParse(dataParser.name, out FileAttachmentAnnotationIcon icon))
                {
                    annotation.Icon = icon;
                }
            }
            return annotation;
        }

        private static AnnotationBase UpdateAnnotation(SoundAnnotation annotation, DataParser dataParser)
        {
            _UpdateMarkupAnnotation(annotation, dataParser);
            var data = dataParser.annotationData;
            bool remove_content = dataParser.ParseBool("remove_content", false);
            if (remove_content)
            {
                annotation.Sound = null;
            }
            else
            {
                string file_encoded_content = data.ContainsKey("file_encoded_content") ? data["file_encoded_content"].ToString() : string.Empty;
                if (!string.IsNullOrEmpty(file_encoded_content))
                {
                    byte[] bytes = System.Convert.FromBase64String(file_encoded_content);
                    annotation.Sound = SoundObject.FromStream(EmbeddedFileStream.FromBytes(dataParser.page.Doc, bytes).GetStream());
                }
            }
            if (!string.IsNullOrEmpty(dataParser.name))
            {
                annotation.Name = dataParser.name;
                if (!string.IsNullOrEmpty(dataParser.name))
                {
                    annotation.Icon = dataParser.name;
                }
            }
            return annotation;
        }

        private static AnnotationBase UpdateAnnotation(CircleAnnotation annotation, DataParser dataParser)
        {
            _UpdateMarkupAnnotation(annotation, dataParser);
            if (dataParser.borderWidth.HasValue)
                annotation.LineWidth = dataParser.borderWidth.Value;
            if (dataParser.lineDashPattern != null)
                annotation.LineDashPattern = dataParser.lineDashPattern;
            if (dataParser.borderStyleType.HasValue)
            {
                if (dataParser.borderStyleType.Value != BorderStyle.Dashed)
                    annotation.LineDashPattern = null;
            }
            annotation.FillColor = dataParser.ParseColor("interiorColor", Color.Transparent).Value;
            return annotation;
        }

        private static AnnotationBase UpdateAnnotation(LineAnnotation annotation, DataParser dataParser)
        {
            _UpdateMarkupAnnotation(annotation, dataParser);
            if (dataParser.borderWidth.HasValue)
                annotation.LineWidth = dataParser.borderWidth.Value;
            if (dataParser.lineDashPattern != null)
                annotation.LineDashPattern = dataParser.lineDashPattern;
            
            if (dataParser.TryParseFloatArray("lineCoordinates", out float[] lineCoordinates))
            {
                annotation.Start = new PointF(lineCoordinates[0], dataParser.ChangeOriginFromBottomToTop(lineCoordinates[1]));
                annotation.End = new PointF(lineCoordinates[2], dataParser.ChangeOriginFromBottomToTop(lineCoordinates[3]));
            }
            if (dataParser.borderStyleType.HasValue)
            {
                if (dataParser.borderStyleType.Value != BorderStyle.Dashed)
                    annotation.LineDashPattern = null;
            }
            if (dataParser.annotationData.ContainsKey("lineStart"))
            {
                if (Enum.TryParse(dataParser.annotationData["lineStart"], out LineEndingStyle lineStartStyle))
                {
                    annotation.LineStartStyle = lineStartStyle;
                }
            }
            if (dataParser.annotationData.ContainsKey("lineEnd"))
            {
                if (Enum.TryParse(dataParser.annotationData["lineEnd"], out LineEndingStyle lineEndStyle))
                {
                    annotation.LineEndStyle = lineEndStyle;
                }
            }
            annotation.LineEndingsFillColor = dataParser.ParseColor("interiorColor", Color.Transparent).Value;
            return annotation;
        }

        private static AnnotationBase UpdateAnnotation(SquareAnnotation annotation, DataParser dataParser)
        {
            _UpdateMarkupAnnotation(annotation, dataParser);
            if (dataParser.borderWidth.HasValue)
                annotation.LineWidth = dataParser.borderWidth.Value;
            if (dataParser.lineDashPattern != null)
                annotation.LineDashPattern = dataParser.lineDashPattern;
            if (dataParser.borderStyleType.HasValue)
            {
                if (dataParser.borderStyleType.Value != BorderStyle.Dashed)
                    annotation.LineDashPattern = null;
            }
            annotation.FillColor = dataParser.ParseColor("interiorColor", Color.Transparent).Value;
            return annotation;
        }

        private static AnnotationBase UpdateAnnotation(RedactAnnotation annotation, DataParser dataParser)
        {
            _UpdateMarkupAnnotation(annotation, dataParser);

            annotation.MarkBorderColor = dataParser.ParseColor("markBorderColor");
            annotation.MarkFillColor = dataParser.ParseColor("markFillColor");
            annotation.OverlayFillColor = dataParser.ParseColor("overlayFillColor", Color.Transparent).Value;
            annotation.OverlayText = dataParser.ParseString("overlayText");
            annotation.OverlayTextRepeat = dataParser.ParseBool("overlayTextRepeat");
            if (Enum.TryParse(dataParser.ParseString("overlayTextJustification"), out VariableTextJustification align)) 
            {
                annotation.Justification = align;
            }
            return annotation;
        }
        
        private static AnnotationBase UpdateAnnotation(TextAnnotation annotation, DataParser dataParser)
        {
            _UpdateMarkupAnnotation(annotation, dataParser);
            if (!string.IsNullOrEmpty(dataParser.name))
            {
                annotation.Name = dataParser.name;
                if (Enum.TryParse(dataParser.name, out TextAnnotationIcon textAnnotationIcon))
                {
                    annotation.Icon = textAnnotationIcon;
                }
            }            
            annotation.Open = dataParser.open;
            return annotation;
        }

        private static AnnotationBase UpdateAnnotation(FreeTextAnnotation annotation, DataParser dataParser)
        {
            _UpdateMarkupAnnotation(annotation, dataParser);
            if (dataParser.borderWidth.HasValue)
                annotation.LineWidth = dataParser.borderWidth.Value;
            if (dataParser.lineDashPattern != null)
                annotation.LineDashPattern = dataParser.lineDashPattern;
            if (dataParser.borderStyleType.HasValue)
            {
                if (dataParser.borderStyleType.Value != BorderStyle.Dashed)
                {
                    annotation.LineDashPattern = null;
                }
            }
            if (dataParser.TryParseFloatArray("calloutLine", out float[] array))
            {
                List<PointF> calloutPoints = new List<PointF>();
                for (int i = 0; i < array.Length; i += 2)
                {
                    if (array.Length > i + 1)
                    {
                        var p = new PointF(array[i], dataParser.ChangeOriginFromBottomToTop(array[i + 1]));
                        calloutPoints.Add(p);
                    }
                }
                annotation.CalloutLine = calloutPoints.ToArray();
            }
            if (dataParser.annotationData.ContainsKey("calloutLineEnd"))
            {
                if (Enum.TryParse(dataParser.annotationData["calloutLineEnd"], out LineEndingStyle calloutLineEndStyle))
                {
                    annotation.LineEndStyle = calloutLineEndStyle;
                }
            }
            if (Enum.TryParse(dataParser.ParseString("textAlignment"), out VariableTextJustification justification))
            {
                annotation.Justification = justification;
            }
            //todo: annotation.LineEndFillColor = dataParser.ParseColor("interiorColor");
            int fontSize = dataParser.ParseNumber("fontSize", 12);
            Color appearanceColor = dataParser.ParseColor("appearanceColor", Color.Empty).Value;
            string colorStr = DataParser.ColorToPdfString(appearanceColor);
            annotation.DefaultAppearanceString = $"{(string.IsNullOrEmpty(colorStr) ? "" : $"{colorStr} ")}{fontSize} Tf";
            return annotation;
        }

        private static AnnotationBase UpdateAnnotation(InkAnnotation annotation, DataParser dataParser)
        {
            _UpdateMarkupAnnotation(annotation, dataParser);
            if (dataParser.borderWidth.HasValue)
                annotation.LineWidth = dataParser.borderWidth.Value;
            if (dataParser.lineDashPattern != null)
                annotation.LineDashPattern = dataParser.lineDashPattern;
            if (dataParser.annotationData.ContainsKey("inkLists"))
            {

                JArray inkLists = dataParser.annotationData["inkLists"];
                foreach (var inkList1 in inkLists)
                {
                    var inkList = inkList1.ToObject<JArray>();
                    List<PointF> path = new List<PointF>();
                    foreach (JObject point in inkList)
                    {
                        float x = float.Parse(point.GetValue("x").ToString());
                        float y = float.Parse(point.GetValue("y").ToString());
                        path.Add(new PointF(x, dataParser.ChangeOriginFromBottomToTop(y)));
                    }
                    annotation.Paths.Add(path.ToArray());
                }
                
            }
            if (dataParser.borderStyleType.HasValue)
            {
                if (dataParser.borderStyleType.Value != BorderStyle.Dashed)
                    annotation.LineDashPattern = null;
            }
            return annotation;
        }

        private static void _UpdatePolygonAnnotationBase(PolygonAnnotationBase annotation, DataParser dataParser)
        {
            if (dataParser.borderWidth.HasValue)
                annotation.LineWidth = dataParser.borderWidth.Value;
            if (dataParser.lineDashPattern != null)
                annotation.LineDashPattern = dataParser.lineDashPattern;
            annotation.FillColor = dataParser.ParseColor("interiorColor", Color.Transparent).Value;
            if (dataParser.annotationData.ContainsKey("vertices"))
            {
                JArray vertices = dataParser.annotationData["vertices"];
                List<PointF> path = new List<PointF>();
                foreach (JObject point in vertices)
                {
                    float x = float.Parse(point.GetValue("x").ToString());
                    float y = float.Parse(point.GetValue("y").ToString());
                    path.Add(new PointF(x, dataParser.ChangeOriginFromBottomToTop(y)));
                }
                annotation.Points = path;
            }
            if (dataParser.borderStyleType.HasValue)
            {
                if (dataParser.borderStyleType.Value != BorderStyle.Dashed)
                    annotation.LineDashPattern = null;
            }
        }

        private static AnnotationBase UpdateAnnotation(PolyLineAnnotation annotation, DataParser dataParser)
        {
            _UpdateMarkupAnnotation(annotation, dataParser);
            _UpdatePolygonAnnotationBase(annotation, dataParser);
            if (dataParser.annotationData.ContainsKey("lineStart"))
            {
                if (Enum.TryParse(dataParser.annotationData["lineStart"], out LineEndingStyle lineStartStyle))
                {
                    annotation.LineStartStyle = lineStartStyle;
                }
            }
            if (dataParser.annotationData.ContainsKey("lineEnd"))
            {
                if (Enum.TryParse(dataParser.annotationData["lineEnd"], out LineEndingStyle lineEndStyle))
                {
                    annotation.LineEndStyle = lineEndStyle;
                }
            }
            annotation.LineEndingsFillColor = dataParser.ParseColor("interiorColor", Color.Transparent).Value;
            return annotation;
        }

        private static AnnotationBase UpdateAnnotation(PolygonAnnotation annotation, DataParser dataParser)
        {
            _UpdateMarkupAnnotation(annotation, dataParser);
            _UpdatePolygonAnnotationBase(annotation, dataParser);
            // BorderEffect
            return annotation;
        }


        #endregion
    }

}
