﻿using GcPdfViewerSupportApi.Models;
using GrapeCity.Documents.Pdf;
using GrapeCity.Documents.Pdf.AcroForms;
using GrapeCity.Documents.Pdf.Annotations;
using GrapeCity.Documents.Pdf.Writer;
using Newtonsoft.Json.Linq;
using System;
using System.Collections.Generic;
using System.Drawing;
using System.Globalization;
using System.Linq;
using System.Text;

namespace GcPdfViewerSupportApi.Utils
{
    public class DataParser
    {

        public RectangleF rect;
        public string title; // (user name)
        public string contents;
        public bool isRichContents;
        public bool open;
        public string subject;
        public string name;
        public Color color;
        public float? borderWidth;
        public BorderStyle? borderStyleType;
        public float[] lineDashPattern;
        public Dictionary<string, dynamic> annotationData;
        public Page page;
        public RectangleF pageBounds;

        /// <summary>
        /// Annotation object which was created or updated by current data parser.
        /// </summary>
        public AnnotationBase Annotation { get; internal set; }

        /// <summary>
        /// Field object which was created or updated by current data parser.
        /// </summary>
        public Field Field { get; internal set; }

        public string id
        {
            get
            {
                if (!annotationData.ContainsKey("id"))
                    return "";
                return annotationData["id"];
            }
        }

        public PdfObjID pdfObjID
        {
            get
            {
                return IdToPdfObjID(id);
            }
        }

        /// <summary>
        /// Order index. Annotations should be sorted according order index. Value of -1 means order is not specified.
        /// </summary>
        public int orderIndex
        {
            get
            {
                return ParseNumber("orderIndex", -1);
            }
        }

        public AnnotationTypeCode annotationType
        {
            get
            {
                if (annotationData.ContainsKey("annotationType"))
                    return (AnnotationTypeCode)int.Parse(annotationData["annotationType"].ToString());
                else
                    return AnnotationTypeCode.UNKNOWN;
            }
        }

        public bool IsWidget
        {
            get
            {
                var annotationType = this.annotationType;
                return annotationType == AnnotationTypeCode.WIDGET || annotationType == AnnotationTypeCode.SIGNATURE;
            }
        }
        public bool IsRadioButton
        {
            get
            {
                return ParseString("fieldType") == "Btn" && ParseBool("radioButton");
            }
        }

        public string FieldName
        {
            get { 
                return ParseString("fieldName");
            }
        }


        public static PdfObjID IdToPdfObjID(string annotationId)
        {
            if (string.IsNullOrEmpty(annotationId))
                return PdfObjID.Empty;
            if (int.TryParse(annotationId.ToUpperInvariant().Replace("R", ""), out int result))
            {
                return new PdfObjID(result, 0);
            }
            return PdfObjID.Empty;
        }

        public DataParser(dynamic annotation, Page page)
        {
            Dictionary<string, dynamic> annotationData = annotation.annotation;
            this.annotationData = annotationData;
            this.page = page;
            pageBounds = page.Bounds; ;
            RectangleF rect;
            if (annotationData.ContainsKey("rect"))
            {
                JArray arr = annotationData["rect"];
                int x1 = arr[0].ToObject<int>(), y1 = arr[1].ToObject<int>(), x2 = arr[2].ToObject<int>(), y2 = arr[3].ToObject<int>();
                rect = new RectangleF(x1, ChangeOriginFromBottomToTop(y1 + (y2 - y1)), x2 - x1, y2 - y1);
            }
            else
            {
                rect = new RectangleF(0, 0, 0, 0);
            }
            this.rect = rect;
            title = annotationData.ContainsKey("title") ? annotationData["title"].ToString() : string.Empty;
            contents = annotationData.ContainsKey("contents") ? annotationData["contents"].ToString() : string.Empty;
            isRichContents = ParseBool("isRichContents");
            open = ParseBool("open");
            subject = ParseString("subject");
            name = ParseString("name");
            color = ParseColor("color", Color.Transparent).Value;
            float? borderWidth = null;
            BorderStyle? borderStyleType = null;
            float[] lineDashPattern = null;
            if (annotationData.ContainsKey("borderStyle"))
            {
                var borderStyle = annotationData["borderStyle"] as JObject;
                if (borderStyle != null)
                {
                    if (borderStyle.ContainsKey("width"))
                    {
                        if (float.TryParse(borderStyle["width"].ToString(), out var borderWidthVal))
                        {
                            borderWidth = borderWidthVal;
                        }

                    }
                    if (borderStyle.ContainsKey("style"))
                    {
                        if (int.TryParse(borderStyle["style"].ToString(), out var borderStyleTypeVal))
                        {
                            if (borderStyleTypeVal < 0 || borderStyleTypeVal > 6)
                                borderStyleType = BorderStyle.Unknown;
                            else
                                borderStyleType = (BorderStyle)borderStyleTypeVal;
                        }
                    }

                    if (TryParseFloatArray(borderStyle, "dashArray", out float[] dashArrayVal))
                    {
                        lineDashPattern = dashArrayVal;
                    }

                }
            }
            this.borderWidth = borderWidth;
            this.borderStyleType = borderStyleType;
            this.lineDashPattern = lineDashPattern;
        }

        public float ChangeOriginFromBottomToTop(float top)
        {
            RectangleF pageBounds = this.pageBounds;
            return (pageBounds.Y + pageBounds.Height) - top + pageBounds.Y;
        }

        public bool ParseBool( string key, bool defaultValue = false)
        {
            Dictionary<string, dynamic> data = this.annotationData;
            if (data.ContainsKey(key))
            {
                if (data[key] is bool)
                    return (bool)data[key];
                else
                {
                    return $"{data[key]}".ToBoolean(defaultValue);
                }
            }
            return defaultValue;
        }

        internal static void CollectAnnotationsData(out List<DataParser> annotationsToUpdate, out List<DataParser> annotationsToCreate,
                        out List<KeyValuePair<int, PdfObjID>> annotationToRemove, out List<DataParser> allClientAnnotations, DocumentModifications _modifications, GcPdfDocument _doc)
        {
            annotationsToUpdate = new List<DataParser>();
            annotationsToCreate = new List<DataParser>();
            allClientAnnotations = new List<DataParser>();
            annotationToRemove = new List<KeyValuePair<int, PdfObjID>>();

            var updatedAnnotations = _modifications.annotationsData.updatedAnnotations;
            var newAnnotations = _modifications.annotationsData.newAnnotations;
            if (updatedAnnotations != null)
            {
                foreach (var updatedAnnotation in updatedAnnotations)
                {
                    Page page = _doc.Pages[updatedAnnotation.pageIndex];
                    var dataParser = new DataParser(updatedAnnotation, page);
                    annotationsToUpdate.Add(dataParser);
                }
            }
            if (newAnnotations != null)
            {
                foreach (var newAnnotation in newAnnotations)
                {
                    var page = _doc.Pages[newAnnotation.pageIndex];
                    var dataParser = new DataParser(newAnnotation, page);
                    annotationsToCreate.Add(dataParser);
                }
            }
            allClientAnnotations.AddRange(annotationsToUpdate);
            allClientAnnotations.AddRange(annotationsToCreate);

            var removedAnnotations = _modifications.annotationsData.removedAnnotations;
            if (removedAnnotations != null)
            {
                foreach (var removedAnnotation in removedAnnotations)
                {
                    if (!string.IsNullOrEmpty(removedAnnotation.annotationId))
                    {
                        var removedAnnotationId = removedAnnotation.annotationId;
                        PdfObjID objID = DataParser.IdToPdfObjID(removedAnnotationId);
                        annotationToRemove.Add(new KeyValuePair<int, PdfObjID>(removedAnnotation.pageIndex, objID));
                    }
                }
            }
        }

        public string ParseString(string key, string defaultValue = "")
        {
            return annotationData.ContainsKey(key) ? (annotationData[key] != null ? annotationData[key].ToString() : defaultValue) : defaultValue;
        }

        public string ParseString(JObject obj, string key, string defaultValue = "")
        {
            if (obj == null)
                return defaultValue;
            return obj.ContainsKey(key) ? (obj[key] != null ? obj[key].ToString() : defaultValue) : defaultValue;
        }

        public int ParseNumber(string key, int defaultValue = 0)
        {
            return annotationData.ContainsKey(key) ? (annotationData[key] != null ? int.Parse(annotationData[key].ToString()) : defaultValue) : defaultValue;
        }

        public bool TryParseArray(string key, out JToken[] arrayValue)
        {
            if (!annotationData.ContainsKey(key))
            {
                arrayValue = null;
                return false;
            }
            return TryParseArrayInternal(annotationData[key], out arrayValue);
        }

        public bool TryParseArray(JObject dictionary, string key, out JToken[] arrayValue)
        {
            if (!dictionary.ContainsKey(key))
            {
                arrayValue = null;
                return false;
            }
            return TryParseArrayInternal(dictionary[key], out arrayValue);
        }

        public bool TryParseFloatArray(string key, out float[] arrayValue)
        {
            if (!annotationData.ContainsKey(key))
            {
                arrayValue = null;
                return false;
            }
            return TryParseFloatArrayInternal(annotationData[key], out arrayValue);
        }

        public bool TryParseFloatArray(JObject dictionary, string key, out float[] arrayValue)
        {
            if (!dictionary.ContainsKey(key))
            {
                arrayValue = null;
                return false;
            }
            return TryParseFloatArrayInternal(dictionary[key], out arrayValue);
        }

        public bool TryParseColor(string key, out Color color)
        {
            var c = ParseColor(key);
            if (c.HasValue)
            {
                color = c.Value;
                return true;
            }
            return false;
        }

        public Color? ParseColor(string key, Color? defaultValue = null)
        {

            dynamic obj = annotationData.ContainsKey(key) ? annotationData[key] : null;
            if (obj == null)
                return defaultValue;
            JArray colorArr = obj as JArray;
            if (colorArr != null)
            {
                int r = int.Parse(colorArr[0].ToString());
                int g = int.Parse(colorArr[1].ToString());
                int b = int.Parse(colorArr[2].ToString());
                int a = colorArr.Count > 3 ? (int)float.Parse(colorArr[3].ToString(), CultureInfo.InvariantCulture) * 255 : -1;
                if (a != -1)
                    return Color.FromArgb(a, r, g, b);
                else
                    return Color.FromArgb(r, g, b);
            }
            JObject colorObj = obj as JObject;
            if (colorObj != null)
            {
                int r = int.Parse(colorObj.GetValue("0").ToString());
                int g = int.Parse(colorObj.GetValue("1").ToString());
                int b = int.Parse(colorObj.GetValue("2").ToString());
                int a = -1;
                if (colorObj.TryGetValue("3", out var token))
                    a = (int)float.Parse(colorObj.GetValue("2").ToString(), CultureInfo.InvariantCulture) * 255;
                if (a != -1)
                    return Color.FromArgb(a, r, g, b);
                else
                    return Color.FromArgb(r, g, b);
            }
            Color? hexColor = ColorUtil.HexToColor(obj.ToString());
            return hexColor ?? defaultValue;

        }

        #region ** private implementations

        private bool TryParseFloatArrayInternal(JToken obj, out float[] arrayValue)
        {
            try
            {
                if (obj == null)
                {
                    arrayValue = null;
                    return false;
                }
                JArray arr = obj as JArray;
                if (arr != null)
                {
                    arrayValue = new float[arr.Count];
                    for (int i = 0; i < arr.Count; i++)
                        arrayValue[i] = float.Parse(arr[i].ToString(), CultureInfo.InvariantCulture);
                    return true;
                }
                JObject arrObj = obj as JObject;
                if (arrObj != null)
                {
                    arrayValue = new float[arrObj.Count];
                    int i = 0;
                    while (i < arrayValue.Length)
                    {
                        if (arrObj.ContainsKey(i.ToString()))
                        {
                            arrayValue[i] = float.Parse(arrObj[i.ToString()].ToString(), CultureInfo.InvariantCulture);
                        }
                        else
                        {
                            arrayValue[i] = 0;
                        }
                        i++;
                    }
                    return true;
                }
                arrayValue = null;
                return false;
            }
            catch (Exception)
            {
                arrayValue = null;
                return false;
            }
        }

        private bool TryParseArrayInternal(JToken obj, out JToken[] arrayValue)
        {
            try
            {
                if (obj == null)
                {
                    arrayValue = null;
                    return false;
                }
                JArray arr = obj as JArray;
                if (arr != null)
                {
                    arrayValue = new JToken[arr.Count];
                    for (int i = 0; i < arr.Count; i++)
                        arrayValue[i] = arr[i];
                    return true;
                }
                JObject arrObj = obj as JObject;
                if (arrObj != null)
                {
                    arrayValue = new JToken[arrObj.Count];
                    int i = 0;
                    while (i < arrayValue.Length)
                    {
                        if (arrObj.ContainsKey(i.ToString()))
                        {
                            arrayValue[i] = arrObj[i.ToString()];
                        }
                        else
                        {
                            arrayValue[i] = null;
                        }
                        i++;
                    }
                    return true;
                }
                arrayValue = null;
                return false;
            }
            catch (Exception)
            {
                arrayValue = null;
                return false;
            }
        }

        internal static string ColorToPdfString(Color color)
        {
            if (color.IsEmpty)
                return "";
            return $"{(color.R / 255f).ToString(CultureInfo.InvariantCulture)} {(color.G / 255f).ToString(CultureInfo.InvariantCulture)} {(color.B / 255f).ToString(CultureInfo.InvariantCulture)} rg";
        }

        #endregion

    }

}
