﻿using GrapeCity.Documents.Pdf;
using System;
using System.Collections.Generic;

using System.IO;
using System.Linq;

using _Float = System.Single;
using _Size = System.Drawing.SizeF;

using GcPdfViewerSupportApi.Models;
using GrapeCity.Documents.Pdf.Annotations;
using GrapeCity.Documents.Pdf.AcroForms;

namespace GcPdfViewerSupportApi.Utils
{

    /// <summary>
    /// PDF document loader.
    /// </summary>
    public class GcPdfDocumentLoader : IDisposable
    {
        #region ** fields

        private static object lockDocumentLoad = new object();
        private _Float _resolution;
        private Stream _stream;
        private GcPdfDocument _doc;

        private DocumentModifications _modifications;
        private SignatureInfo _signatureInfo;
        private SupportApiSettings _settings;

        #endregion

        #region ** constructor

        /// <summary>
        /// Constructor.
        /// </summary>
        public GcPdfDocumentLoader(SupportApiSettings settings, _Float resolution = 96)
        {
            _settings = settings;
            _resolution = resolution;
            Url = "";
            DocId = string.Format("_{0:N}", Guid.NewGuid());
            
        }

        #endregion

        #region ** properties

        public string DocId { get; private set; }

        public OpenedDocumentInfo Info { get; private set; }

        public string Url { get; }

        public int Rotate { get; set; } = 0;

        public _Float Scale { get; set; } = 1.0f;

        public bool RenderInteractiveForms { get; internal set; }

        #endregion

        #region ** methods

        public OpenedDocumentInfo Open(string url)
        {
            using (var client = new WebClientWithTimeout())
            {
                var bytes = client.DownloadData(url);
                var data = new MemoryStream(bytes);
                return Open(data);
            }
        }

        public OpenedDocumentInfo Open(Stream data)
        {
            lock (lockDocumentLoad)
            {
                if (_stream != null)
                {
                    _stream.Dispose();
                    _stream = null;
                    _doc = null;
                }
                _stream = data;
                return _OpenInitialDocState();

            }
        }

        public void Reset()
        {
            _modifications = null;
            _signatureInfo = null;
        }

        public void ApplyDocumentModifications(DocumentModifications modifications)
        {
            _modifications = modifications;
        }

        public void Sign(SignatureInfo signatureInfo)
        {
            _signatureInfo = signatureInfo;
        }

        public void Save(MemoryStream s)
        {
            _applyModifications();                        
            if (_signatureInfo != null)
            {
                _saveWithSignature(s);
            }
            else
            {
                _doc.Save(s);
            }
            s.Seek(0, SeekOrigin.Begin);
        }

        public bool VerifySignature(string fieldName)
        {
            var signatureField = _doc.AcroForm.Fields.Where(f => f.Name == fieldName).FirstOrDefault() as SignatureField;
            if(signatureField == null || signatureField.Value == null)
            {
                return false;
            }
            return signatureField.Value.VerifySignature();
        }

        public void Dispose()
        {
            _doc = null;
            if (_stream != null)
            {
                _stream.Dispose();
                _stream = null;
            }
        }

        #endregion

        #region ** private and internal implementation

        private void _applyModifications()
        {

            _OpenInitialDocState();// Reset previous changes.
            if (_modifications != null)
            {
                if (_modifications.rotation.HasValue)
                {
                    Rotate = _modifications.rotation.Value;
                }
                if (_modifications.renderInteractiveForms.HasValue)
                {
                    RenderInteractiveForms = _modifications.renderInteractiveForms.Value;
                }
                // Apply annotation modifications
                if (_modifications.annotationsData != null)
                {
                    List<KeyValuePair<PopupAnnotation, string>> popupAnnotations = new List<KeyValuePair<PopupAnnotation, string>>();

                    DataParser.CollectAnnotationsData(out List<DataParser> annotationsToUpdate, out List<DataParser> annotationsToCreate,
                        out List<KeyValuePair<int, PdfObjID>> annotationToRemove, out List<DataParser> allClientAnnotations, _modifications, _doc);
                    foreach (var dataParser in annotationsToUpdate)
                    {                        
                        PdfObjID objID = dataParser.pdfObjID;
                        AnnotationTypeCode annotationType = dataParser.annotationType;
                        try
                        {
                            if (dataParser.IsWidget)
                            {
                                var field = _doc.AcroForm.Fields.Where(a => a.ParsedObjID == objID).FirstOrDefault();
                                if (field == null)
                                {
                                    if (dataParser.IsRadioButton)
                                        field = _doc.AcroForm.Fields.Where(a => a.Widgets.Any(w => w.ParsedObjID == objID)).FirstOrDefault();
                                    if (field == null)
                                        continue;
                                }
                                AcroFormFactory.UpdateField(annotationType, field, dataParser);
                                AddFieldWithSort(field, _doc.AcroForm.Fields, dataParser, allClientAnnotations);
                            }
                            else
                            {
                                AnnotationBase annotation = dataParser.page.Annotations.Where(a => a.ParsedObjID == objID).FirstOrDefault();
                                if (annotation == null)
                                {
                                    //throw new Exception($"Annotation with id {objID} not found.");
                                    continue;
                                }
                                AnnotationFactory.UpdateAnnotation(annotationType, annotation, dataParser);
                                AddAnnotationWithSort(annotation, dataParser.page.Annotations, dataParser, allClientAnnotations);
                                if (annotationType == AnnotationTypeCode.REDACT && dataParser.ParseBool("redactApplied"))
                                {
                                    _doc.Redact(annotation as RedactAnnotation, _settings.RedactOptions);
                                }
                                if(annotationType == AnnotationTypeCode.POPUP)
                                {
                                    popupAnnotations.Add(new KeyValuePair<PopupAnnotation, string>((PopupAnnotation)annotation, dataParser.ParseString("parentId")));
                                }
                            }


                        }
                        catch (Exception) {
                            continue;
                        }
                    }

                    foreach (var removedAnnotation in annotationToRemove)
                    {
                        try
                        {
                            Page page = _doc.Pages[removedAnnotation.Key];
                            PdfObjID objID = removedAnnotation.Value;
                            var annotations = page.Annotations;
                            var annotation = annotations.Where(a => a.ParsedObjID == objID).FirstOrDefault();
                            if (annotation != null)
                            {
                                annotations.Remove(annotation);
                            }
                            else
                            {
                                var field = _doc.AcroForm.Fields.Where(a => a.ParsedObjID == objID).FirstOrDefault();
                                if (field != null)
                                {
                                    _doc.AcroForm.Fields.Remove(field);
                                }
                            }
                        }
                        catch (Exception) { 
                            continue; 
                        }
                    }

                    foreach (var dataParser in annotationsToCreate)
                    {
                        try
                        {
                            var page = dataParser.page;
                            if (dataParser.IsWidget)
                            {
                                if (dataParser.IsRadioButton)
                                {
                                    string fieldName = dataParser.FieldName;
                                    var radioButtonFieldWithSameName = _doc.AcroForm.Fields.Where(a => a.Name == fieldName).FirstOrDefault();
                                    if (radioButtonFieldWithSameName != null)
                                    {
                                        var widget = new WidgetAnnotation(dataParser.page, dataParser.rect);
                                        radioButtonFieldWithSameName.Widgets.Add(widget);
                                        AcroFormFactory.UpdateFieldWidget(radioButtonFieldWithSameName, widget, dataParser);
                                        // TODO: sort widgets
                                        //AddFieldWithSort(radioButtonFieldWithSameName, _doc.AcroForm.Fields, dataParser, allClientAnnotations);
                                        continue;
                                    }
                                }
                                Field field = AcroFormFactory.CreateField(dataParser, page);
                                if (field != null)
                                {
                                    AddFieldWithSort(field, _doc.AcroForm.Fields, dataParser, allClientAnnotations);
                                }
                            }
                            else
                            {
                                AnnotationBase annotation = AnnotationFactory.CreateAnnotation(dataParser, page);
                                if (annotation != null)
                                {
                                    AddAnnotationWithSort(annotation, page.Annotations, dataParser, allClientAnnotations);
                                    if (dataParser.annotationType == AnnotationTypeCode.REDACT && dataParser.ParseBool("redactApplied"))
                                    {
                                        _doc.Redact(annotation as RedactAnnotation, _settings.RedactOptions);
                                    }
                                    if (dataParser.annotationType == AnnotationTypeCode.POPUP)
                                    {
                                        popupAnnotations.Add(new KeyValuePair<PopupAnnotation, string>((PopupAnnotation)annotation, dataParser.ParseString("parentId")));
                                    }
                                }
                            }
                        }
                        catch (Exception) {
                            continue;
                        }
                    }

                    // Associate popup annotations with their destinations:
                    foreach (var keyValue in popupAnnotations)
                    {
                        try
                        {
                            var popupAnnotation = keyValue.Key;
                            var parentId = keyValue.Value;
                            var parentObjId = DataParser.IdToPdfObjID(parentId);

                            MarkupAnnotation referencedMarkupAnnotation = popupAnnotation.Page.Annotations.Where(a => a.ParsedObjID == parentObjId).FirstOrDefault() as MarkupAnnotation;
                            if(referencedMarkupAnnotation == null)
                            {
                                var annotationParser = allClientAnnotations.Where(p => p.id == parentId).FirstOrDefault();
                                if (annotationParser != null)
                                    referencedMarkupAnnotation = annotationParser.Annotation as MarkupAnnotation;
                            }
                            if (referencedMarkupAnnotation != null)
                            {
                                referencedMarkupAnnotation.Popup = popupAnnotation;
                            }

                        }
                        catch (Exception)
                        {
                            continue;
                        }
                        
                    }

                }
                // Import form data
                if (_modifications.formData != null && _modifications.formData.Count > 0)
                {
                    try
                    {
                        List<KeyValuePair<string, IList<string>>> fieldValues = new List<KeyValuePair<string, IList<string>>>();
                        foreach (var pageIndex in _modifications.formData.Keys)
                        {
                            if (_modifications.formData.TryGetValue(pageIndex, out Dictionary<string, string[]> pageData))
                            {
                                foreach (var kv in pageData)
                                {
                                    fieldValues.Add(new KeyValuePair<string, IList<string>>(kv.Key, new List<string>(kv.Value)));
                                }
                            }
                        }
                        _doc.ImportFormDataFromCollection(fieldValues.ToArray());
                    } 
                    catch (Exception)
                    {
                        //
                    }
                }
            }
            int rotate = Rotate;
            var pages = _doc.Pages;
            foreach (var p in pages)
            {
                p.Rotate = rotate;
            }

        }

        private void AddFieldWithSort(Field field, FieldCollection collection, DataParser dataParser, List<DataParser> allClientAnnotations)
        {
            AddFieldWithSort(field, collection, new List<DataParser>() { dataParser }, allClientAnnotations);
        }

        private void AddFieldWithSort(Field field, FieldCollection collection, List<DataParser> dataParsers, List<DataParser> allClientAnnotations)
        {
            if (collection.Contains(field))
                collection.Remove(field);

            int orderIndex = -1;
            Page page = null;            
            foreach (var dataParser in dataParsers)
            {
                dataParser.Field = field;
                orderIndex = dataParser.orderIndex;
                page = dataParser.page;
                break;
            }
            
            
            if (orderIndex == -1)
            {
                collection.Add(field);
            }
            else
            {
                var sorted = false;
                try
                {
                    var insertAfterRef = allClientAnnotations.Where(p => p.orderIndex < orderIndex && p.pdfObjID != field.ParsedObjID).OrderBy(p => p.orderIndex).FirstOrDefault();
                    if (insertAfterRef != null)
                    {
                        var refAnnot = collection.Where(a => a.ParsedObjID == insertAfterRef.pdfObjID).FirstOrDefault();
                        if (refAnnot != null)
                        {
                            int insertIndex = collection.IndexOf(refAnnot);
                            collection.Insert(insertIndex + 1, field);
                            sorted = true;
                        }
                    }
                    else
                    {
                        var insertBeforeRef = allClientAnnotations.Where(p => p.orderIndex > orderIndex && p.pdfObjID != field.ParsedObjID).OrderByDescending(p => p.orderIndex).FirstOrDefault();
                        if (insertBeforeRef != null)
                        {
                            var refAnnot = collection.Where(a => a.ParsedObjID == insertBeforeRef.pdfObjID).FirstOrDefault();
                            if (refAnnot != null)
                            {
                                int insertIndex = collection.IndexOf(refAnnot);
                                collection.Insert(insertIndex, field);
                                sorted = true;
                            }
                        }
                    }

                }
                catch (Exception)
                {

                }
                if (!sorted)
                {
                    collection.Add(field);
                }
            }
            foreach (var widget in field.Widgets)
            {
                widget.Page = page; 
            }
        }

        private void AddAnnotationWithSort(AnnotationBase annotation, AnnotationCollection collection, DataParser dataParser, List<DataParser> allClientAnnotations)
        {
            if(collection.Contains(annotation))
                collection.Remove(annotation);
            dataParser.Annotation = annotation;
            
            int orderIndex = dataParser.orderIndex;
            if (orderIndex == -1)
            {
                collection.Add(annotation);
            }
            else
            {
                var sorted = false;
                try
                {
                    var insertAfterRef = allClientAnnotations.Where(p => p.orderIndex < orderIndex && p.pdfObjID != annotation.ParsedObjID).OrderBy(p => p.orderIndex).FirstOrDefault();
                    if (insertAfterRef != null)
                    {
                        var refAnnot = collection.Where(a => a.ParsedObjID == insertAfterRef.pdfObjID).FirstOrDefault();
                        if (refAnnot != null)
                        {
                            int insertIndex = collection.IndexOf(refAnnot);
                            collection.Insert(insertIndex + 1, annotation);
                            sorted = true;
                        }
                    }
                    else
                    {
                        var insertBeforeRef = allClientAnnotations.Where(p => p.orderIndex > orderIndex && p.pdfObjID != annotation.ParsedObjID).OrderByDescending(p => p.orderIndex).FirstOrDefault();
                        if (insertBeforeRef != null)
                        {
                            var refAnnot = collection.Where(a => a.ParsedObjID == insertBeforeRef.pdfObjID).FirstOrDefault();
                            if (refAnnot != null)
                            {
                                int insertIndex = collection.IndexOf(refAnnot);
                                collection.Insert(insertIndex, annotation);
                                sorted = true;
                            }
                        }
                    }
                } 
                catch (Exception)
                {

                }
                if (!sorted)
                {
                    collection.Add(annotation);
                }
            }
            annotation.Page = dataParser.page;
        }

        private void _saveWithSignature(Stream stream)
        {            
            if (_signatureInfo != null)
            {
                
                var props = new SignatureProperties();
                props.ContactInfo = _signatureInfo.ContactInfo;
                props.Location = _signatureInfo.Location;
                props.SignerName = _signatureInfo.SignerName;
                props.SigningDateTime = DateTime.Now;
                props.Reason = _signatureInfo.Reason;
                props.SignatureDigestAlgorithm = _signatureInfo.SignatureDigestAlgorithm;
                props.SignatureFormat = _signatureInfo.SignatureFormat;
                props.TimeStamp = _signatureInfo.TimeStamp;
                props.SignatureField = _signatureInfo.SignatureField;
                props.Certificate = _settings.Certificate;
                _doc.Sign(props, stream);
            }
        }

        private OpenedDocumentInfo _OpenInitialDocState()
        {
            if (_doc != null)
            {
                _doc.Clear();
                _doc = null;
            }
            var d = new GcPdfDocument();
            d.Load(_stream);
            _doc = d;
            _Size defaultViewPortSize = d.Pages.Count > 0 ? d.Pages[0].GetRenderSize(this._resolution, this._resolution) : new _Size(0f, 0f);
            Info = new OpenedDocumentInfo(d.DocumentInfo, d.Pages.Count, DocId, defaultViewPortSize);
            return Info;
        }

#endregion

    }
}
